#coding: utf-8

from json import loads, dumps

from os.path import splitext

from thumbor.storages import BaseStorage

from ..aws.storage import AwsStorage


class Storage(AwsStorage, BaseStorage):

    def __init__(self, context):
        BaseStorage.__init__(self, context)
        AwsStorage.__init__(self, context, 'STORAGE')

    def put(self, path, bytes):
        self.set(bytes, self.normalize_path(path))

        return path

    def put_crypto(self, path):
        if not self.context.config.STORES_CRYPTO_KEY_FOR_EACH_IMAGE:
            return

        if not self.context.server.security_key:
            raise RuntimeError("STORES_CRYPTO_KEY_FOR_EACH_IMAGE can't be True if no SECURITY_KEY specified")

        file_abspath = self.normalize_path(path)
        crypto_path = '%s.txt' % splitext(file_abspath)[0]

        self.set(self.context.server.security_key, crypto_path)

        return crypto_path

    def put_detector_data(self, path, data):
        file_abspath = self.normalize_path(path)

        path = '%s.detectors.txt' % splitext(file_abspath)[0]

        self.set(dumps(data), path)

        return path

    def get_crypto(self, path):
        file_abspath = self.normalize_path(path)
        crypto_path = "%s.txt" % (splitext(file_abspath)[0])

        file_key = self.storage.get_key(crypto_path)

        if not file_key:
            return None

        return file_key.read()

    def get_detector_data(self, path):
        file_abspath = self.normalize_path(path)
        path = '%s.detectors.txt' % splitext(file_abspath)[0]

        file_key = self.storage.get_key(path)

        if not file_key or self.is_expired(file_key):
            return None

        return loads(file_key.read())

    def exists(self, path):
        file_abspath = self.normalize_path(path)
        file_key = self.storage.get_key(file_abspath)

        if not file_key:
            return False

        return True

    def remove(self, path):
        if not self.exists(path):
            return

        if not self.storage.delete_key(path):
            return False

        return True

    def resolve_original_photo_path(self, filename):
        return filename



# coding: utf-8

from boto.s3.bucket import Bucket
from thumbor.utils import logger
from tornado.concurrent import return_future

import thumbor.loaders.http_loader as http_loader

from tc_aws.loaders import *
from tc_aws.aws.connection import get_connection


def _generate_presigned_url(context, bucket, key):
    connection = get_connection(context)
    expiry = 60 * 60
    presigned_url = connection.generate_url(
        expiry,
        'GET',
        bucket,
        key,
    )
    return presigned_url


@return_future
def load(context, url, callback):
    load_sync(context, url, callback)


def load_sync(context, url, callback):
    if _use_http_loader(context, url):
        http_loader.load_sync(
            context, url, callback, normalize_url_func=_normalize_url)
    else:
        bucket, key = _get_bucket_and_key(context, url)

        if _validate_bucket(context, bucket):
            presigned_url = _generate_presigned_url(context, bucket, key)
            http_loader.load_sync(
                context, presigned_url, callback, normalize_url_func=_normalize_url)
        else:
            callback(None)



# coding: utf-8

from boto.s3.connection import S3Connection

connection = None


def get_connection(context):
    global connection

    if connection is None:
        boto_opts = {}

        if context.config.AWS_ROLE_BASED_CONNECTION is False:
            boto_opts.update({
                'aws_access_key_id': context.config.AWS_ACCESS_KEY,
                'aws_secret_access_key': context.config.AWS_SECRET_KEY,
                })

        if context.config.BOTO_CONFIG:
            boto_opts.update(context.config.BOTO_CONFIG)

        connection = S3Connection(**boto_opts)

    return connection



# coding: utf-8

from thumbor.config import Config

Config.define('STORAGE_BUCKET',                         'thumbor-images',   'S3 bucket for Storage', 'S3')
Config.define('RESULT_STORAGE_BUCKET',                  'thumbor-result',   'S3 bucket for result Storage', 'S3')
Config.define('S3_LOADER_BUCKET',                       None,               'S3 bucket for loader', 'S3')

Config.define('S3_LOADER_ROOT_PATH',                    '',                 'S3 path prefix for Loader bucket', 'S3')
Config.define('STORAGE_AWS_STORAGE_ROOT_PATH',          '',                 'S3 path prefix for Storage bucket', 'S3')
Config.define('RESULT_STORAGE_AWS_STORAGE_ROOT_PATH',   '',                 'S3 path prefix for Result storage bucket', 'S3')

Config.define('STORAGE_EXPIRATION_SECONDS',             3600,               'S3 expiration', 'S3')

Config.define('S3_STORAGE_SSE',                         False,              'S3 encriptipon key', 'S3')
Config.define('S3_STORAGE_RRS',                         False,              'S3 redundency', 'S3')
Config.define('S3_ALLOWED_BUCKETS',                     False,              'List of allowed bucket to be requeted', 'S3')
Config.define('RESULT_STORAGE_S3_STORE_METADATA',       False,              'S3 store result with metadata', 'S3')

Config.define('AWS_ACCESS_KEY',                         None,               'AWS Access key, if None use environment AWS_ACCESS_KEY_ID', 'AWS')
Config.define('AWS_SECRET_KEY',                         None,               'AWS Secret key, if None use environment AWS_SECRET_ACCESS_KEY', 'AWS')
Config.define('AWS_ROLE_BASED_CONNECTION',              False,              'EC2 instance can use role that does not require AWS_ACCESS_KEY see http://docs.aws.amazon.com/IAM/latest/UserGuide/roles-usingrole-ec2instance.html', 'AWS')

Config.define('BOTO_CONFIG',                            None,               'Additional Boto options for configuring S3 access (see http://boto.readthedocs.org/en/latest/ref/s3.html?highlight=boto.s3.connection.s3connection#boto.s3.connection.S3Connection)')



#se!/usr/bin/python
# -*- coding: utf-8 -*-

from pyvows import Vows, expect

from thumbor.context import Context
from thumbor.config import Config
from fixtures.storage_fixture import IMAGE_URL, IMAGE_BYTES, get_server
import time

from boto.s3.connection import S3Connection
from boto.s3.key import Key
from moto import mock_s3

from tc_aws.storages.s3_storage import Storage

s3_bucket = 'thumbor-images-test'


@Vows.batch
class S3StorageVows(Vows.Context):

    class CanStoreImage(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            bucket = self.conn.create_bucket(s3_bucket)

            thumborId = IMAGE_URL % '1'
            config = Config(STORAGE_BUCKET=s3_bucket)
            storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
            store = storage.put(thumborId, IMAGE_BYTES)
            k = Key(bucket)
            k.key = thumborId
            result = k.get_contents_as_string()
            return (store, result)

        def should_be_in_catalog(self, topic):
            expect(topic[0]).to_equal(IMAGE_URL % '1')
            expect(topic[1]).not_to_be_null()
            expect(topic[1]).not_to_be_an_error()
            expect(topic[1]).to_equal(IMAGE_BYTES)

    class CanGetImage(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(STORAGE_BUCKET=s3_bucket)
            storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
            storage.put(IMAGE_URL % '2', IMAGE_BYTES)
            return storage.get(IMAGE_URL % '2')

        def should_not_be_null(self, topic):
            expect(topic).not_to_be_null()
            expect(topic).not_to_be_an_error()

        def should_have_proper_bytes(self, topic):
            expect(topic).to_equal(IMAGE_BYTES)

    class CanGetImageExistance(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(STORAGE_BUCKET=s3_bucket)
            storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
            storage.put(IMAGE_URL % '3', IMAGE_BYTES)
            return storage.exists(IMAGE_URL % '3')

        def should_exists(self, topic):
            expect(topic).to_equal(True)

    class CanGetImageInexistance(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(STORAGE_BUCKET=s3_bucket)
            storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
            return storage.exists(IMAGE_URL % '9999')

        def should_not_exists(self, topic):
            expect(topic).to_equal(False)

    class CanRemoveImage(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(STORAGE_BUCKET=s3_bucket)
            storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
            storage.put(IMAGE_URL % '4', IMAGE_BYTES)
            created = storage.exists(IMAGE_URL % '4')
            time.sleep(1)
            storage.remove(IMAGE_URL % '4')
            time.sleep(1)
            return storage.exists(IMAGE_URL % '4') != created

        def should_be_put_and_removed(self, topic):
            expect(topic).to_equal(True)

    class CanRemovethenPutImage(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(STORAGE_BUCKET=s3_bucket)
            storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
            storage.put(IMAGE_URL % '5', IMAGE_BYTES)
            storage.remove(IMAGE_URL % '5')
            time.sleep(1)
            created = storage.exists(IMAGE_URL % '5')
            time.sleep(1)
            storage.put(IMAGE_URL % '5', IMAGE_BYTES)
            return storage.exists(IMAGE_URL % '5') != created

        def should_be_put_and_removed(self, topic):
            expect(topic).to_equal(True)

    class CanReturnPath(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(STORAGE_BUCKET=s3_bucket)
            storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
            return storage.resolve_original_photo_path("toto")

        def should_return_the_same(self, topic):
            expect(topic).to_equal("toto")

    class HandlesStoragePrefix(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(STORAGE_BUCKET=s3_bucket, STORAGE_AWS_STORAGE_ROOT_PATH='tata')
            storage = Storage(Context(config=config, server=get_server('ACME-SEC')))

            return storage.normalize_path('toto')

        def should_return_the_same(self, topic):
            expect(topic).to_equal("tata/toto")

    class CryptoVows(Vows.Context):
        class RaisesIfInvalidConfig(Vows.Context):
            @Vows.capture_error
            @mock_s3
            def topic(self):
                self.conn = S3Connection()
                self.conn.create_bucket(s3_bucket)

                config = Config(STORAGE_BUCKET=s3_bucket, STORES_CRYPTO_KEY_FOR_EACH_IMAGE=True)
                storage = Storage(Context(config=config, server=get_server('')))
                storage.put(IMAGE_URL % '9999', IMAGE_BYTES)
                storage.put_crypto(IMAGE_URL % '9999')

            def should_be_an_error(self, topic):
                expect(topic).to_be_an_error_like(RuntimeError)
                expect(topic).to_have_an_error_message_of("STORES_CRYPTO_KEY_FOR_EACH_IMAGE can't be True if no SECURITY_KEY specified")

        class GettingCryptoForANewImageReturnsNone(Vows.Context):
            @mock_s3
            def topic(self):
                self.conn = S3Connection()
                self.conn.create_bucket(s3_bucket)

                config = Config(STORAGE_BUCKET=s3_bucket, STORES_CRYPTO_KEY_FOR_EACH_IMAGE=True)
                storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
                return storage.get_crypto(IMAGE_URL % '9999')

            def should_be_null(self, topic):
                expect(topic).to_be_null()

        class DoesNotStoreIfConfigSaysNotTo(Vows.Context):
            @mock_s3
            def topic(self):
                self.conn = S3Connection()
                self.conn.create_bucket(s3_bucket)

                config = Config(STORAGE_BUCKET=s3_bucket)
                storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
                storage.put(IMAGE_URL % '9998', IMAGE_BYTES)
                storage.put_crypto(IMAGE_URL % '9998')
                return storage.get_crypto(IMAGE_URL % '9998')

            def should_be_null(self, topic):
                expect(topic).to_be_null()

        class CanStoreCrypto(Vows.Context):
            @mock_s3
            def topic(self):
                self.conn = S3Connection()
                self.conn.create_bucket(s3_bucket)

                config = Config(STORAGE_BUCKET=s3_bucket, STORES_CRYPTO_KEY_FOR_EACH_IMAGE=True)
                storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
                storage.put(IMAGE_URL % '6', IMAGE_BYTES)
                storage.put_crypto(IMAGE_URL % '6')
                return storage.get_crypto(IMAGE_URL % '6')

            def should_not_be_null(self, topic):
                expect(topic).not_to_be_null()
                expect(topic).not_to_be_an_error()

            def should_have_proper_key(self, topic):
                expect(topic).to_equal('ACME-SEC')

    class DetectorVows(Vows.Context):
        class CanStoreDetectorData(Vows.Context):
            @mock_s3
            def topic(self):
                self.conn = S3Connection()
                self.conn.create_bucket(s3_bucket)

                config = Config(STORAGE_BUCKET=s3_bucket)
                storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
                storage.put(IMAGE_URL % '7', IMAGE_BYTES)
                storage.put_detector_data(IMAGE_URL % '7', 'some-data')
                return storage.get_detector_data(IMAGE_URL % '7')

            def should_not_be_null(self, topic):
                expect(topic).not_to_be_null()
                expect(topic).not_to_be_an_error()

            def should_equal_some_data(self, topic):
                expect(topic).to_equal('some-data')

        class ReturnsNoneIfNoDetectorData(Vows.Context):
            @mock_s3
            def topic(self):
                self.conn = S3Connection()
                self.conn.create_bucket(s3_bucket)

                config = Config(STORAGE_BUCKET=s3_bucket)
                storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
                return storage.get_detector_data(IMAGE_URL % '9999')

            def should_not_be_null(self, topic):
                expect(topic).to_be_null()



# coding: utf-8
__all__ = ['_get_bucket_and_key', '_get_bucket', '_normalize_url', '_validate_bucket', '_use_http_loader']

import urllib2


def _get_bucket_and_key(context, url):
    url = urllib2.unquote(url)

    bucket = context.config.get('S3_LOADER_BUCKET', default=None)

    if not bucket:
        bucket, key = _get_bucket(url, root_path=context.config.S3_LOADER_ROOT_PATH)
    else:
        key = url

    return bucket, key


def _get_bucket(url, root_path=None):
    url_by_piece = url.lstrip("/").split("/")
    bucket_name = url_by_piece[0]

    if root_path is not None:
        url_by_piece[0] = root_path
    else:
        url_by_piece = url_by_piece[1:]

    bucket_path = "/".join(url_by_piece)

    return bucket_name, bucket_path


def _normalize_url(url):
    """
    :param url:
    :return: exactly the same url since we only use http loader if url stars with http prefix.
    """
    return url


def _validate_bucket(context, bucket):
    allowed_buckets = context.config.get('S3_ALLOWED_BUCKETS', default=None)
    return not allowed_buckets or bucket in allowed_buckets


def _use_http_loader(context, url):
    enable_http_loader = context.config.get('AWS_ENABLE_HTTP_LOADER', default=False)
    return enable_http_loader and url.startswith('http')



#coding: utf-8

from thumbor.result_storages import BaseStorage

from ..aws.storage import AwsStorage


class Storage(AwsStorage, BaseStorage):
    def __init__(self, context):
        BaseStorage.__init__(self, context)
        AwsStorage.__init__(self, context, 'RESULT_STORAGE')

    def put(self, bytes):
        path = self.normalize_path(self.context.request.url)
        self.set(bytes, path)

        return path

    def get(self, path=None):
        if path is None:
            path = self.normalize_path(self.context.request.url)

        return super(Storage, self).get(path)



# coding: utf-8

from boto.s3.bucket import Bucket
from thumbor.utils import logger
from tornado.concurrent import return_future

import thumbor.loaders.http_loader as http_loader

from tc_aws.loaders import *
from tc_aws.aws.connection import get_connection


@return_future
def load(context, url, callback):
    load_sync(context, url, callback)

def load_sync(context, url, callback):
    if _use_http_loader(context, url):
        return http_loader.load_sync(context, url, callback, normalize_url_func=_normalize_url)

    bucket, key = _get_bucket_and_key(context, url)

    if _validate_bucket(context, bucket):
        bucket_loader = Bucket(
            connection=get_connection(context),
            name=bucket
        )
        file_key = None
        try:
            file_key = bucket_loader.get_key(url)
        except Exception, e:
            logger.warn("ERROR retrieving image from S3 {0}: {1}".format(url, str(e)))

        if file_key:
            callback(file_key.read())
            return

    callback(None)



#se!/usr/bin/python
# -*- coding: utf-8 -*-

from pyvows import Vows, expect

from thumbor.context import Context
from tc_aws import Config
from fixtures.storage_fixture import IMAGE_BYTES, get_server

from boto.s3.connection import S3Connection
from moto import mock_s3

from tc_aws.result_storages.s3_storage import Storage

s3_bucket = 'thumbor-images-test'


class Request(object):
    url = None


@Vows.batch
class S3ResultStorageVows(Vows.Context):

    class CanStoreImage(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(RESULT_STORAGE_BUCKET=s3_bucket)
            ctx = Context(config=config, server=get_server('ACME-SEC'))
            ctx.request = Request
            ctx.request.url = 'my-image.jpg'

            storage = Storage(ctx)
            path = storage.put(IMAGE_BYTES)

            return path

        def should_be_in_catalog(self, topic):
            expect(topic).to_equal('my-image.jpg')

    class CanGetImage(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(RESULT_STORAGE_BUCKET=s3_bucket)
            ctx = Context(config=config, server=get_server('ACME-SEC'))
            ctx.request = Request
            ctx.request.url = 'my-image-2.jpg'

            storage = Storage(ctx)
            storage.put(IMAGE_BYTES)

            return storage.get()

        def should_not_be_null(self, topic):
            expect(topic).not_to_be_null()
            expect(topic).not_to_be_an_error()

        def should_have_proper_bytes(self, topic):
            expect(topic).to_equal(IMAGE_BYTES)

    class CanGetImageWithMetadata(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(RESULT_STORAGE_BUCKET=s3_bucket, RESULT_STORAGE_S3_STORE_METADATA=True)
            ctx = Context(config=config, server=get_server('ACME-SEC'))
            ctx.headers = {'Content-Type': 'image/webp', 'Some-Other-Header': 'doge-header'}
            ctx.request = Request
            ctx.request.url = 'my-image-meta.jpg'

            storage = Storage(ctx)
            storage.put(IMAGE_BYTES)

            file_abspath = storage.normalize_path(ctx.request.url)
            file_key = storage.storage.get_key(file_abspath)
            return file_key.content_type, file_key.metadata, file_key.read()

        def should_have_proper_bytes(self, topic):
            expect(topic[0]).to_include('image/webp')
            expect(topic[1]).to_include('some-other-header')
            expect(topic[2]).to_equal(IMAGE_BYTES)

    class HandlesStoragePrefix(Vows.Context):
        @mock_s3
        def topic(self):
            self.conn = S3Connection()
            self.conn.create_bucket(s3_bucket)

            config = Config(RESULT_STORAGE_BUCKET=s3_bucket, RESULT_STORAGE_AWS_STORAGE_ROOT_PATH='tata')
            ctx = Context(config=config, server=get_server('ACME-SEC'))

            storage = Storage(ctx)

            return storage.normalize_path('toto')

        def should_return_the_same(self, topic):
            expect(topic).to_equal("tata/toto")



# se!/usr/bin/python
# -*- coding: utf-8 -*-

from urlparse import urlparse, parse_qs

from mock import Mock

from pyvows import Vows, expect
from mock import patch

from thumbor.context import Context
from derpconf.config import Config

import boto
from boto.s3.key import Key

from moto import mock_s3

from fixtures.storage_fixture import IMAGE_PATH, IMAGE_BYTES

from tc_aws.loaders import presigning_loader

s3_bucket = 'thumbor-images-test'

@Vows.batch
class PresigningLoaderVows(Vows.Context):

    class CanLoadImage(Vows.Context):

        @mock_s3
        def topic(self):
            conn = boto.connect_s3()
            bucket = conn.create_bucket(s3_bucket)

            k = Key(bucket)
            k.key = IMAGE_PATH
            k.set_contents_from_string(IMAGE_BYTES)

            conf = Config()
            conf.define('S3_LOADER_BUCKET', s3_bucket, '')
            conf.define('S3_LOADER_ROOT_PATH', 'root_path', '')

            return Context(config=conf)

        def should_load_from_s3(self, topic):
            image = yield presigning_loader.load(topic, '/'.join(['root_path', IMAGE_PATH]))
            expect(image).to_equal(IMAGE_BYTES)

    class ValidatesBuckets(Vows.Context):

        @mock_s3
        def topic(self):
            conf = Config()
            conf.define('S3_ALLOWED_BUCKETS', [], '')

            return Context(config=conf)

        def should_load_from_s3(self, topic):
            image = yield presigning_loader.load(topic, '/'.join([s3_bucket, IMAGE_PATH]))
            expect(image).to_equal(None)

    class HandlesHttpLoader(Vows.Context):

        @mock_s3
        def topic(self):
            conf = Config()
            conf.define('AWS_ENABLE_HTTP_LOADER', True, '')

            return Context(config=conf)

        @patch('thumbor.loaders.http_loader.load_sync')
        def should_redirect_to_http(self, topic, load_sync_patch):
            def callback(*args):
                pass

            presigning_loader.load_sync(topic, 'http://foo.bar', callback)
            expect(load_sync_patch.called).to_be_true()

    class CanBuildPresignedUrl(Vows.Context):

        def topic(self):
            conf = Config()
            return Context(config=conf)

        @mock_s3
        def should_generate_presigned_urls(self, topic):
            url = presigning_loader._generate_presigned_url(
                topic, "bucket-name", "some-s3-key")
            url = urlparse(url)
            expect(url.scheme).to_equal('https')
            expect(url.hostname).to_equal('bucket-name.s3.amazonaws.com')
            expect(url.path).to_equal('/some-s3-key')
            url_params = parse_qs(url.query)
            # We can't test Expires & Signature values as they vary depending on the TZ
            expect(url_params).to_include('Expires')
            expect(url_params).to_include('Signature')
            expect(url_params['AWSAccessKeyId'][0]).to_equal('test-key')
            expect(url_params['x-amz-security-token'][0]).to_equal('test-session-token')



# coding: utf-8

import calendar

from os.path import join

from datetime import datetime, timedelta

from thumbor.utils import logger

from boto.s3.bucket import Bucket
from boto.s3.key import Key
from dateutil.parser import parse as parse_ts

from connection import get_connection


class AwsStorage():

    @property
    def is_auto_webp(self):
        return self.context.config.AUTO_WEBP and self.context.request.accepts_webp

    def __init__(self, context, config_prefix):
        self.config_prefix = config_prefix
        self.context = context
        self.storage = self.__get_s3_bucket()

    def _get_config(self, config_key, default=None):
        return getattr(self.context.config, '%s_%s' % (self.config_prefix, config_key))

    def __get_s3_bucket(self):
        return Bucket(
            connection=get_connection(self.context),
            name=self._get_config('BUCKET')
        )

    def set(self, bytes, abspath):
        file_key = Key(self.storage)
        file_key.key = abspath

        if self.config_prefix is 'RESULT_STORAGE' and self._get_config('S3_STORE_METADATA'):
            for k, v in self.context.headers.iteritems():
                file_key.set_metadata(k, v)

        file_key.set_contents_from_string(
            bytes,
            encrypt_key=self.context.config.get('S3_STORAGE_SSE', ''),         # TODO: fix config prefix
            reduced_redundancy=self.context.config.get('S3_STORAGE_RRS', '')   # TODO: fix config prefix
        )

    def get(self, path):
        file_abspath = self.normalize_path(path)

        file_key = self.storage.get_key(file_abspath)

        if not file_key or self.is_expired(file_key):
            logger.debug("[STORAGE] s3 key not found at %s" % file_abspath)
            return None
        return file_key.read()

    def normalize_path(self, path):
        path = path.lstrip('/')  # Remove leading '/'
        path_segments = [self._get_config('AWS_STORAGE_ROOT_PATH'), path]

        if self.is_auto_webp:
            path_segments.append("webp")

        return join(*path_segments)

    def is_expired(self, key):
        if key:
            expire_in_seconds = self._get_config('EXPIRATION_SECONDS')

            # Never expire
            if expire_in_seconds is None or expire_in_seconds == 0:
                return False

            timediff = datetime.now() - self.utc_to_local(parse_ts(key.last_modified))
            return timediff.seconds > self._get_config('EXPIRATION_SECONDS')
        else:
            #If our key is bad just say we're expired
            return True

    def last_updated(self):
        path = self.context.request.url
        file_abspath = self.normalize_path(path)
        file_key = self.storage.get_key(file_abspath)

        if not file_key or self.is_expired(file_key):
            logger.debug("[RESULT_STORAGE] s3 key not found at %s" % file_abspath)
            return None

        return self.utc_to_local(parse_ts(file_key.last_modified))

    def utc_to_local(self, utc_dt):
        # get integer timestamp to avoid precision lost
        timestamp = calendar.timegm(utc_dt.timetuple())
        local_dt = datetime.fromtimestamp(timestamp)
        assert utc_dt.resolution >= timedelta(microseconds=1)
        return local_dt.replace(microsecond=utc_dt.microsecond)
